unit MainUnit;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, ExtCtrls,
  AIOAIO;

const
  AM_DEPTH = WM_USER + $97;
  AM_MESSAGE = WM_USER + $98;
  AM_GOTDATA = WM_USER + $99;

type
  TModelType = (AIOProto, AIO, ADIO, HMAIO);

  TMainForm = class(TForm)
    ControlPanel: TPanel;
    GroupBox1: TGroupBox;
    LogMemo: TMemo;
    DeviceCombo: TComboBox;
    HealthTime: TTimer;
    HealthLabel: TLabel;
    Panel3: TPanel;
    DifferentialCheck: TCheckBox;
    Panel2: TPanel;
    HzLabel: TLabel;
    HzEdit: TEdit;
    GoButton: TButton;
    StopButton: TButton;
    LogFormatRadios: TRadioGroup;
    Panel1: TPanel;
    Label2: TLabel;
    AFEdit: TEdit;
    StopDepthCheck: TCheckBox;
    StopDepthEdit: TEdit;
    DepthLabel: TLabel;
    DataList: TListBox;
    Label3: TLabel;
    RangePanelADAS0: TPanel;
    RangeRadios0: TRadioGroup;
    RangeRadios1: TRadioGroup;
    RangeRadios2: TRadioGroup;
    RangeRadios3: TRadioGroup;
    RangeRadios4: TRadioGroup;
    RangeRadios5: TRadioGroup;
    RangeRadios6: TRadioGroup;
    RangeRadios7: TRadioGroup;
    RangePanelADAS1: TPanel;
    RangeRadios8: TRadioGroup;
    RangeRadios9: TRadioGroup;
    RangeRadios10: TRadioGroup;
    RangeRadios11: TRadioGroup;
    RangeRadios12: TRadioGroup;
    RangeRadios13: TRadioGroup;
    RangeRadios14: TRadioGroup;
    RangeRadios15: TRadioGroup;
    ChannelCount: TComboBox;
    ScanGroup: TGroupBox;
    ScanHzEdit: TEdit;
    Label5: TLabel;
    Label6: TLabel;
    Label4: TLabel;
    ScanCheck: TCheckBox;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure GoButtonClick(Sender: TObject);
    procedure StopButtonClick(Sender: TObject);
    procedure DataListKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure HealthTimeTimer(Sender: TObject);
    procedure DifferentialCheckClick(Sender: TObject);
    procedure DeviceComboChange(Sender: TObject);
    procedure ScanCheckClick(Sender: TObject);
  private
    function UpdateChannelsAvailable: integer;
    function DetectModel(DeviceID: TUInt32): TModelType;
  protected
    bPollStarts: Boolean;
    LastData: Int64;
    RunStart, LastDataChange: Int64;
    RunLength: Single;
    RunsKnown: Integer;
    StopDepth: Int64;
    StartStopCmd: TUInt32;
    StartBitMask: TUInt32;

    RangeRadios: array[0..16] of TRadioGroup;

    procedure EnableGo(bEnable: Boolean);
    procedure EnableStop(bEnable: Boolean);
    procedure DoCycleRun(NewHz: Double; ThisRunLength: Int64);
    procedure ProcessPause;
    procedure AMGotData(var Msg: TMessage); message AM_GOTDATA;
    procedure AMMessage(var Msg: TMessage); message AM_MESSAGE;
    procedure AMDepth(var Msg: TMessage); message AM_DEPTH;
    { Private declarations }
  public
    { Public declarations }
  end;


var
  Model: TModelType;
  Version:Longword;
  bScanMode: Boolean;
  MainForm: TMainForm;

implementation

{$R *.DFM}

uses
  Math,
  ShellAPI, Clipbrd;

const
  ERROR_TRANSLATION_COMPLETE = 757;

function LocalAlloc(uFlags, uBytes: UINT): Pointer; stdcall; external kernel32;
function LocalFree(pMem: Pointer): Pointer; stdcall; external kernel32;

function PIO_DMAPiongControl(iBoard: TBoardIndex; pControlReply: TPUInt8; ControlSize: TUInt32; ReplySize: TPUInt32 = nil): TWinErrCode; cdecl; external 'AIOAIO.dll';
function PIO_DMAPiongIn(iBoard: TBoardIndex; pData: TPUInt8; DataSize: TUInt32; pOverlap: POverlapped): TWinErrCode; cdecl; external 'AIOAIO.dll';

function PIO_GetDeviceHandle(iBoard: TBoardIndex; var hDevice: THandle): TWinErrCode; cdecl; external 'AIOAIO.dll';
//function PIO_QueryBARBase(iBoard: TBoardIndex; iBAR: LongWord; pBase: PLongWord): TWinErrCode; cdecl; external 'AIOAIO.dll';

function DoDMAPiongControlReply(iBoard: TBoardIndex; var ControlReply: AnsiString): TWinErrCode;
var
  Size: TUInt32;

begin
  Size := Length(ControlReply);
  Result := PIO_DMAPiongControl(iBoard, TPUint8(@ControlReply[1]), Size, @Size);
  SetLength(ControlReply, Size);
end;

function DoDMAPiongControl(iBoard: TBoardIndex; Control: AnsiString): TWinErrCode;
begin
  Result := DoDMAPiongControlReply(iBoard, Control);
end;

const
	ResetOffset = $00;
	IRQEnableOffset_AIO = $01;
	DACOffset = $04;
	BaseClockOffset = $0C;
	DivisorOffset = $10;
	ADCRangeOffset = $18;
	AFLevelOffset = $20;
	ADCDepthOffset = $28;
	ADCDataOffset: array[0..1] of TUInt16 = ($30, $34);
	ADCStartStopOffset = $38;
	ADCImmediateOffset = $3A;
	IRQEnableOffset_ADIO = $40;
	//ADCDataOffset: array[0..1] of TUInt32 = ($4000, $8000);
	dSequencer = $4;

	BaseImmediateCmd = $8CCE;
	BaseStartStopCmd = $8CEE;

var
  bDualSequencer: Boolean;
  bDifferential: Boolean;

// 2,4,6,8,10,12,14,16 = AIO S.E.
// 2,4,6,8 = AIO Diff
// 1..8 = ADIO S.E.
// 1..4 = ADIO Diff
function ItemIndexToCount(Index: TUInt8): TUInt8;
begin
  Result := Index + 1;
  if bDualSequencer then
    Result := Result * 2; // mPCIe-AIO always takes pairs-of-channels
end;

function NybbleToHex(N: Byte): AnsiChar;
begin
  N := N and $F;
  if N < 10 then
    Result := AnsiChar(Byte('0') + N)
  else
    Result := AnsiChar(Byte('A') - 10 + N)
  ;
end;

function HexToNybble(Hex: AnsiChar): Byte;
begin
  case Hex of
  '0'..'9': Result := Byte(Hex) - Byte('0');
  'A'..'F': Result := Byte(Hex) - Byte('A') + 10;
  'a'..'f': Result := Byte(Hex) - Byte('a') + 10;
  else      raise Exception.CreateFmt('Invalid hex character "%s"', [ Hex ]);
  end;
end;

function StrToHex(Text: AnsiString): AnsiString;
var
  I: Integer;

begin
  SetString(Result, nil, Length(Text) * 2);
  for I := 1 to Length(Text) do begin
    Result[I * 2 - 1] := NybbleToHex(Byte(Text[I]) shr 4);
    Result[I * 2]     := NybbleToHex(Byte(Text[I])      );
  end;
end;

function HexToStr(Hex: AnsiString): AnsiString;
var
  I: Integer;

begin
  SetString(Result, nil, (Length(Hex) + 1) div 2);
  for I := 1 to Length(Result) do
    Result[I] := AnsiChar(HexToNybble(Hex[I * 2]) + HexToNybble(Hex[I * 2 - 1]) shl 4)
  ;
end;

function StrToSpacedHex(Text: AnsiString): AnsiString;
var
  I: Integer;

begin
  Result := AnsiString(StringOfChar(' ', Length(Text) * 3-1));
  for I := 1 to Length(Text) do begin
    Result[I * 3 - 2] := NybbleToHex(Byte(Text[I]) shr 4);
    Result[I * 3 - 1] := NybbleToHex(Byte(Text[I])      );
  end;
end;

function StrToSpacedWords(Text: AnsiString): AnsiString;
var
  TI, RI, L: Integer;

begin
  L := Length(Text);
  Result := AnsiString(StringOfChar(' ', (L shr 1)*5 + (L and 1)*3 - 1));
  TI := 1;
  RI := 1;
  while L >= 2 do begin
    Inc(TI);
    Result[RI] := NybbleToHex(Byte(Text[TI]) shr 4);
    Inc(RI);
    Result[RI] := NybbleToHex(Byte(Text[TI])      );
    Inc(RI); Dec(TI);
    Result[RI] := NybbleToHex(Byte(Text[TI]) shr 4);
    Inc(RI);
    Result[RI] := NybbleToHex(Byte(Text[TI])      );
    Inc(RI, 2); Inc(TI, 2); Dec(L, 2);
  end;

  if L > 0 then begin
    Result[RI] := NybbleToHex(Byte(Text[TI]) shr 4);
    Inc(RI);
    Result[RI] := NybbleToHex(Byte(Text[TI])      );
    //Inc(RI, 2); Inc(TI); Dec(L);
  end;
end;

procedure DumpStringToFile(S, Path: AnsiString);
var
  Fil: TFileStream;

begin
  Fil := TFileStream.Create(String(Path), fmCreate or fmShareDenyWrite);
  try
    Fil.Write(S[1], Length(S));
  finally
    Fil.Free;
  end;
end;

function DumpFileToString(Path: AnsiString): AnsiString;
var
  Fil: TFileStream;
  L: Integer;

begin
  Fil := TFileStream.Create(String(Path), fmOpenRead or fmShareDenyNone);
  try
    L := Fil.Size;
    SetString(Result, nil, L);
    Fil.Read(Result[1], L);
  finally
    Fil.Free;
  end;
end;

function BStatus(bSucc: Boolean): LongWord;
begin
	if bSucc then
		Result := ERROR_SUCCESS
	else begin
		Result := GetLastError;
		if Result = ERROR_SUCCESS then Result := ERROR_GEN_FAILURE;
	end;
end;



type
  TLogThread = class;

  TReadData = record
    pData: Pointer;
    Ovl: TOverlapped;
  end;
  TPReadData = ^TReadData;

  TDataThread = class(TThread)
  public
    constructor Create; virtual;
    destructor Destroy; override;
  public
    hDataWnd: THandle;
    DataMsg: LongWord;
    FNSuffix: AnsiString;
  protected
    function GetDone: Boolean; virtual;
  public property bDone: Boolean read GetDone;
  protected
    hDoneEvent: THandle;
    LogThread: TLogThread;

    procedure Execute; override;
  end;

  TLogThread = class(TThread)
  public
    hDepthWnd: THandle;
    //CachedDepth: Integer;

    constructor Create; virtual;
    destructor Destroy; override;

    procedure OpenFN(Suffix: String); virtual;
    procedure PutItem(NewItem: Pointer); virtual;
    function GetItem: Pointer; virtual;
  protected
    hItemMutex, hItemSem: THandle;
    Items: TList;
    LogFil: TFileStream;

    procedure Execute; override;
  end;

var
  TicksPerSec: Int64;
  DataThread: TDataThread;
  iBoard: TBoardIndex;
  DMADataSize: TUInt32;
  LogFormat: (lfNone, lfBinary, lfSingleSample, lfAllChannels, lfClearAll);
  LastFragmentTicks: Int64;

procedure DisposeRead(pRead: TPReadData);
begin
  if pRead.pData <> nil then begin
    LocalFree(pRead.pData);
    pRead.pData := nil;
  end;
  if pRead.Ovl.hEvent <> 0 then begin
    CloseHandle(pRead.Ovl.hEvent);
    pRead.Ovl.hEvent := 0;
  end;
  Dispose(pRead);
end;

{ TDataThread }

constructor TDataThread.Create;
begin
  inherited Create(True);

  hDoneEvent := CreateEvent(nil, True, False, nil);
end;

destructor TDataThread.Destroy;
var
  H: THandle;

begin
  H := hDoneEvent;
  hDoneEvent := 0;
  CloseHandle(H);

  inherited;
end;

procedure TDataThread.Execute;
var
  Status: TWinErrCode;
  Overlaps, UnusedOverlaps: TList;
  pThisRead: TPReadData;
  iRead: Integer;
  hDevice: THandle;
  DummyBytesTransferred: Longword;
  iCh: Integer;
  //iData: Integer;
  //pData: PLongWord;
  LastData: array of LongWord;

  procedure PostMessageStr(Msg: AnsiString);
  var
    pMsg: PAnsiString;

  begin
    if hDataWnd = INVALID_HANDLE_VALUE then Exit;

    New(pMsg);
    pMsg^ := Msg;
    PostMessage(hDataWnd, AM_MESSAGE, 0, Integer(pMsg));
  end;

begin
  ReturnValue := PIO_GetDeviceHandle(iBoard, hDevice);
  if ReturnValue <> ERROR_SUCCESS then begin
    PostMessageStr('Error ' + AnsiString(IntToStr(ReturnValue)) + ' getting device handle.');
    Exit;
  end;

  LogThread := TLogThread.Create;
  LogThread.hDepthWnd := hDataWnd;
  LogThread.OpenFN(String(FNSuffix));
  LogThread.Start;
  Overlaps := TList.Create;
  UnusedOverlaps := TList.Create;
  SetLength(LastData, $10);
  for iCh := 0 to High(LastData) do LastData[iCh] := $80000000 or TUInt32($F - iCh);

  SetThreadPriority(GetCurrentThread, THREAD_PRIORITY_TIME_CRITICAL);
  repeat
    if Overlaps.Count < 4 then begin
      //PostMessageStr('Overlaps.Count == ' + AnsiString(IntToStr(Overlaps.Count)) + '.');
      if UnusedOverlaps.Count > 0 then begin
        pThisRead := UnusedOverlaps[0];
        UnusedOverlaps.Delete(0);
        //PostMessageStr('UnusedOverlaps.Count == ' + AnsiString(IntToStr(UnusedOverlaps.Count)) + '.');
      end
      else begin
        New(pThisRead);
        ZeroMemory(pThisRead, SizeOf(pThisRead^));
        pThisRead.pData := LocalAlloc(LMEM_FIXED or LMEM_ZEROINIT, DMADataSize);
        pThisRead.Ovl.hEvent := CreateEvent(nil, False, False, nil);
      end;
      Status := PIO_DMAPiongIn(iBoard, pThisRead.pData, DMADataSize, @pThisRead.Ovl);
      case Status of
        ERROR_SUCCESS: ;
        ERROR_IO_PENDING: ;
        else Break;
      end;

      Overlaps.Add(pThisRead);
      pThisRead := nil;
    end
    else begin
      pThisRead := Overlaps[0];
      Overlaps.Delete(0);
      //PostMessageStr('Overlaps.Delete(0);');

      Status := BStatus(GetOverlappedResult(hDevice, pThisRead.Ovl, DummyBytesTransferred, True));
      case Status of
        ERROR_NO_MORE_FILES,
        ERROR_TRANSLATION_COMPLETE: Status := ERROR_SUCCESS;
      end;
      if Status <> ERROR_SUCCESS then Break;

      QueryPerformanceCounter(LastFragmentTicks);

      {
      pData := pThisRead.pData;
      for iData := 0 to (DMADataSize div 4)-1 do begin
        if (pData^ and $80000000) = 0 then begin
          iCh := (pData^ shr 20) and $7;
          if (iData and 1) <> 0 then iCh := iCh or $8;
          LastData[iCh] := pData^;
        end;
        Inc(pData);
      end;

      if (hDataWnd <> INVALID_HANDLE_VALUE) and (DataMsg <> WM_NULL) then
        for iCh := 0 to High(LastData) do
          PostMessage(hDataWnd, DataMsg, iCh, LastData[iCh])
      ;
      }

      LogThread.PutItem(pThisRead.pData);
      pThisRead.pData := LocalAlloc(LMEM_FIXED or LMEM_ZEROINIT, DMADataSize);

      UnusedOverlaps.Add(pThisRead);
      pThisRead := nil;
    end;

    if Terminated then begin
      Status := ERROR_OPERATION_ABORTED;
      Break;
    end;
  until False;
  SetThreadPriority(GetCurrentThread, THREAD_PRIORITY_NORMAL);

  if pThisRead <> nil then begin
    DisposeRead(pThisRead);
    //pThisRead := nil;
  end;

  for iRead := Overlaps.Count-1 downto 0 do DisposeRead(Overlaps[iRead]);
  Overlaps.Free;
  for iRead := UnusedOverlaps.Count-1 downto 0 do DisposeRead(UnusedOverlaps[iRead]);
  UnusedOverlaps.Free;

  LogThread.Free;
  Sleep(1);
  PostMessage(hDataWnd, AM_DEPTH, 0, 0);

  ReturnValue := Status;
  PostMessageStr('Worker thread exiting with ' + AnsiString(IntToStr(ReturnValue)) + '.');
  SetEvent(hDoneEvent);
end;

function TDataThread.GetDone: Boolean;
var
  H: THandle;

begin
  Result := False;

  if Self = nil then Exit;

  H := hDoneEvent;
  if H = 0 then Exit;

  if WaitForSingleObject(hDoneEvent, 0) <> WAIT_OBJECT_0 then Exit;

  Result := True;
end;

{ TLogThread }

procedure TLogThread.OpenFN(Suffix: String);
var
  ST: TSystemTime;
  FN: AnsiString;
  HeadLine: AnsiString;

begin
  GetLocalTime(ST);
  with ST do FN := AnsiString(Format('%.4d%.2d%.2d_%.2d%.2d%.2d ', [ wYear, wMonth, wDay, wHour, wMinute, wSecond ]) + Suffix);
  case LogFormat of
    lfNone: begin
      LogFil := nil;
      Exit;
    end;
    lfBinary: begin
      FN := FN + ' Raw.bin';
      HeadLine := '';
    end;
    lfSingleSample: begin
      FN := FN + ' Single.csv';
      HeadLine := '"Ch OK", "Raw","Channel","Gain","Top Byte","Counts"'#13#10;
    end;
    lfAllChannels: begin
      FN := FN + ' All.csv';
      HeadLine := '"Row#","Ch0","Ch1","Ch2","Ch3","Ch4","Ch5","Ch6","Ch7","Ch8","Ch9","ChA","ChB","ChC","ChD","ChE","ChF"'#13#10;
    end;
    lfClearAll: begin
      FN := FN + ' Clear All.txt';
      HeadLine := 'Sample#'#9' 0000 1111 2222 3333 4444 5555 6666 7777 8888 9999 AAAA BBBB CCCC DDDD EEEE FFFF'#13#10;
    end;
  end;
  LogFil := TFileStream.Create(String(FN), fmCreate);
  LogFil.Free;
  LogFil := TFileStream.Create(String(FN), fmOpenWrite or fmShareDenyWrite);

  if HeadLine <> '' then begin
    LogFil.WriteBuffer(HeadLine[1], Length(HeadLine));
    HeadLine := '';
  end;
end;

constructor TLogThread.Create;
begin
  inherited Create(True);

  hItemMutex := CreateMutex(nil, False, nil);
  hItemSem := CreateSemaphore(nil, 0, $0FFFFFFF, nil);

  Items := TList.Create;
end;

destructor TLogThread.Destroy;
begin
  //With no item, this is a "kill release".
  ReleaseSemaphore(hItemSem, 1, nil);

  inherited;

  WaitForSingleObject(hItemMutex, 0);
  FreeAndNil(Items);
  CloseHandle(hItemMutex);
  CloseHandle(hItemSem);

  FreeAndNil(LogFil);
end;

procedure TLogThread.Execute;
var
  pItem, pData: PLongWord;
  iCh, iCh2, iChBack1, iChBack2, iChExpected, dCh2, iData, iSample: Integer;
  LogBlock: TStringList;
  LogStr: AnsiString;
  LogCh: array of Integer;

  procedure ParseCh;
  begin
    iCh := (pData^ shr 20) and $7;
    if bDualSequencer and ((iData and 1) <> 0) then iCh := iCh or $8;
    if iChBack2 < 0 then begin
      dCh2 := 0;
    end
    else begin
      dCh2 := Cardinal(iCh - iChBack2) mod 8;
      if dCh2 <> 1 then dCh2 := 0;
    end;
  end;

  function bAnyChannels: Boolean;
  var
    iCh: Integer;
  
  begin
    Result := True;
    for iCh := 0 to High(LogCh) do
      if LogCh[iCh] <> MAXINT then
        Exit
    ;
    Result := False;
  end;

  procedure LogInvalid(Sample: LongWord);
  begin
    LogBlock.Add(IntToStr(iSample) + ',INVALID,' + IntToHex(Sample and $7FFFFFFF, 1));

    Inc(iSample);
  end;

  procedure LogAllChannels;
  var
    iCh: Integer;
    Buf: AnsiString;

  begin
    Buf := AnsiString(IntToStr(iSample));
    for iCh := 0 to High(LogCh) do
      if LogCh[iCh] = MAXINT then
        Buf := Buf + ','
      else
        Buf := Buf + (',' + AnsiString(IntToStr(LogCh[iCh])))
    ;
    LogBlock.Add(String(Buf));

    Inc(iSample);
    for iCh := 0 to High(LogCh) do LogCh[iCh] := MAXINT;
  end;


  procedure PostMessageStr(Msg: AnsiString);
  var
    pMsg: PAnsiString;

  begin
    New(pMsg);
    pMsg^ := Msg;
    PostMessage(MainForm.Handle, AM_MESSAGE, 0, Integer(pMsg));
  end;

begin
  LogBlock := TStringList.Create;
  if LogFormat = lfAllChannels then begin
    SetLength(LogCh, 16);
    for iCh := 0 to High(LogCh) do LogCh[iCh] := MAXINT;
  end;
  iSample := 0;
  LogStr := '';

  iChBack2 := -2;
  iChBack1 := -2;
  repeat
    pItem := GetItem;
    if pItem = nil then Break;

    if LogFormat = lfNone then
    else begin
      LogBlock.Clear;
      case LogFormat of
        lfBinary: begin
          SetString(LogStr, PAnsiChar(pItem), DMADataSize);
        end;
        lfSingleSample: begin
          pData := pItem;
          for iData := 0 to (DMADataSize div 4)-1 do begin
            if (pData^ and $80000000) <> 0 then begin
              iCh := -2;
              //LogBlock.Add(IntToHex(pData^, 8) + '  INVALID');
              LogBlock.Add('0,"' + IntToHex(pData^, 8) + '",-1');
            end
            else begin
              ParseCh;
              //LogBlock.Add(IntToHex(pData^, 8) + '  Ch' + NybbleToHex(iCh) + ', range ' + NybbleToHex(pData^ shr 16) + ', ' + IntToHex(pData^ shr 24, 2) + ': ' + IntToHex(pData^ and $FFFF, 4) + ' ct');
              LogBlock.Add('' + NybbleToHex(dCh2) + ',"' + IntToHex(pData^, 8) + '",' + IntToStr(iCh) + ',' + IntToStr((pData^ shr 16) and $F) + ',"' + IntToHex(pData^ shr 24, 2) + '",' + IntToStr(SmallInt(pData^ and $FFFF)));
            end;
            iChBack2 := iChBack1;
            iChBack1 := iCh;
            Inc(pData);
          end;
          LogStr := AnsiString(LogBlock.Text);
          LogBlock.Clear;
        end;
        lfAllChannels: begin
          pData := pItem;
          for iData := 0 to (DMADataSize div 4)-1 do begin
            if (pData^ and $80000000) <> 0 then begin
              if bAnyChannels then LogAllChannels;
              LogInvalid(pData^);
            end
            else begin
              ParseCh;
              if LogCh[iCh] <> MAXINT then LogAllChannels;
              LogCh[iCh] := SmallInt(pData^ and $FFFF);
            end;
            Inc(pData);
          end;
          if bAnyChannels then LogAllChannels;
          LogStr := AnsiString(LogBlock.Text);
          LogBlock.Clear;
        end;
        lfClearAll: begin
          pData := pItem;
          for iData := 0 to (DMADataSize div 4)-1 do begin
            LogStr := AnsiString(IntToHex(iSample, 1) + #9);
            if (pData^ and $80000000) <> 0 then begin
              iCh := -2;
            end
            else begin
              ParseCh;
              if iChBack2 < 0 then
                iChExpected := -1
              else
                iChExpected := (iChBack2 and $8) or ((iChBack2+1) and $7)
              ;
              for iCh2 := $0 to $F do
                if iCh2 = iCh then begin
                  if (pData^ and $8000) <> 0 then
                    LogStr := LogStr + ('-' + AnsiString(IntToHex($10000 - (pData^ and $FFFF), 4)) )
                  else
                    LogStr := LogStr + (' ' + AnsiString(IntToHex(pData^ and $FFFF, 4)) )
                  ;
                end
                else if iCh2 = iChExpected then
                  LogStr := LogStr + ' ????'
                else
                  LogStr := LogStr + ' '
              ;
            end;
            LogBlock.Add(String(LogStr));
            Inc(iSample);
            iChBack2 := iChBack1;
            iChBack1 := iCh;
            Inc(pData);
          end;
          LogStr := AnsiString(LogBlock.Text);
          LogBlock.Clear;
        end;
        else begin
          PostMessageStr('Invalid log format #' + AnsiString(IntToStr(Ord(LogFormat))) + '.');
          Terminate;
          LocalFree(pItem); //We should free the others too, TList won't do it for us
          Break;
        end;
      end;
    end;
    LocalFree(pItem);

    if LogStr <> '' then begin
      LogFil.WriteBuffer(LogStr[1], Length(LogStr));
      LogStr := '';
    end;
  until False;

  FreeAndNil(LogBlock);
end;

function TLogThread.GetItem: Pointer;
var
  CachedDepth: Integer;

begin
  Result := nil;
  CachedDepth := -1;

  if WaitForSingleObject(hItemSem, INFINITE) <> WAIT_OBJECT_0 then Exit;

  WaitForSingleObject(hItemMutex, INFINITE);
    if Items.Count <> 0 then begin
      Result := Items[0];
      Items.Delete(0);
      CachedDepth := Items.Count;
    end;
  ReleaseMutex(hItemMutex);

  if CachedDepth >= 0 then PostMessage(hDepthWnd, AM_DEPTH, 0, CachedDepth);
end;

procedure TLogThread.PutItem(NewItem: Pointer);
var
  CachedDepth: Integer;

begin
  WaitForSingleObject(hItemMutex, INFINITE);
    Items.Add(NewItem);
    CachedDepth := Items.Count;
  ReleaseMutex(hItemMutex);

  PostMessage(hDepthWnd, AM_DEPTH, 0, CachedDepth);

  ReleaseSemaphore(hItemSem, 1, nil);
end;




{ TMainForm ---------------------------------------------------------------------------------------------------}

procedure TMainForm.EnableGo(bEnable: Boolean);
var
  I: Integer;
begin
  DeviceCombo.Enabled := bEnable;
  RangePanelADAS0.Enabled := bEnable;
  RangePanelADAS1.Enabled := bEnable;

  for I := 0 to 15 do
    RangeRadios[I].Enabled := bEnable
  ;

  DifferentialCheck.Enabled := bEnable;
  HzEdit.Enabled := bEnable;
  GoButton.Enabled := bEnable;
end;

procedure TMainForm.EnableStop(bEnable: Boolean);
begin
  StopButton.Enabled := bEnable;
end;

procedure TMainForm.FormCreate(Sender: TObject);
var
  I: Integer;
  DeviceStr: AnsiString;
  DeviceStrSize: TUInt32;
  Status: TWinErrCode;

begin
  RangeRadios[0] := RangeRadios0;
  RangeRadios[1] := RangeRadios1;
  RangeRadios[2] := RangeRadios2;
  RangeRadios[3] := RangeRadios3;
  RangeRadios[4] := RangeRadios4;
  RangeRadios[5] := RangeRadios5;
  RangeRadios[6] := RangeRadios6;
  RangeRadios[7] := RangeRadios7;
  RangeRadios[8] := RangeRadios8;
  RangeRadios[9] := RangeRadios9;
  RangeRadios[10] := RangeRadios10;
  RangeRadios[11] := RangeRadios11;
  RangeRadios[12] := RangeRadios12;
  RangeRadios[13] := RangeRadios13;
  RangeRadios[14] := RangeRadios14;
  RangeRadios[15] := RangeRadios15;

  QueryPerformanceFrequency(TicksPerSec);
  IsMultiThread := True;

  DeviceCombo.Align := alTop;
  DifferentialCheck.Align := alTop;
  DeviceCombo.Items.Clear;
  for I := 0 to BRD_GetHighestBoardIndex do begin
    DeviceStrSize := MAX_PATH;
    DeviceStr := AnsiString(StringOfChar(#0, DeviceStrSize));
    Status := BRD_GetName(I, DeviceStrSize, TPAnsiChar(DeviceStr), @DeviceStrSize);
    if Status <> ERROR_SUCCESS then
      DeviceStr := '<Error ' + AnsiString(IntToStr(Status)) + ' from BRD_GetName>'
    else
      SetLength(DeviceStr, DeviceStrSize)
    ;
    DeviceCombo.Items.Add('[' + IntToStr(I) + '] ' + String(DeviceStr));
  end;

  Application.ProcessMessages;

  if DeviceCombo.Items.Count <> 0 then begin
    DeviceCombo.ItemIndex := 0;
    DeviceComboChange(DeviceCombo);
    DifferentialCheckClick(nil);
    ScanCheckClick(nil);
  end;

  ChannelCount.ItemIndex := UpdateChannelsAvailable;
end;

procedure TMainForm.FormDestroy(Sender: TObject);
begin
  //`@
end;

procedure TMainForm.GoButtonClick(Sender: TObject);
const
  TopSpeed = 1000000;
var
  Status: TWinErrCode;
  Hz, ScanHz: Double;
  //Data: AnsiString;
  ChCt, HighCh, MaxCh, ChPerStart: TUInt32;
  DeviceID: LongWord;
  iCh: Integer;
  BaseClock: TUInt32;

begin
  EnableGo(False);

  if Sender <> nil then begin
    RunLength := 60 * TicksPerSec;
    RunsKnown := 0;
  end;

  if Model in [HMAIO, ADIO] then begin // if ADIO or HMAIO
    StartStopCmd := BaseStartStopCmd or $20000;
    StartBitMask := $10000;
  end
  else begin                             // AIO, AIOProto
    StartStopCmd := BaseStartStopCmd;
    StartBitMask := $1;
  end;

  if Model in [HMAIO, AIO, AIOProto] then begin
    MaxCh := 16;
    bDualSequencer := True;
  end else begin
    MaxCh := 8;
    bDualSequencer := False;
  end;

  ChCt := ItemIndexToCount(ChannelCount.ItemIndex);

  if bDualSequencer and ((ChCt and 1) <> 0) then raise Exception.Create('Can''t acquire odd number of channels.');
  if ChCt < (MaxCh div 8) then raise Exception.Create('Can''t acquire ' + IntToStr(ChCt) + ' channels.');//TODO?: modify for HMAIO features
  if bDifferential then begin
    if ChCt > (MaxCh div 2) then raise Exception.Create('Can''t acquire ' + IntToStr(ChCt) + ' channels.');
  end
  else begin
    if ChCt > MaxCh then raise Exception.Create('Can''t acquire ' + IntToStr(ChCt) + ' channels.');
  end;

  if bDualSequencer then ChCt := ChCt div 2;
  HighCh := ChCt - 1;

  Hz := StrToFloat(HzEdit.Text);

  if bScanMode then begin
    ScanHz := StrToFloat(ScanHzEdit.Text);
    if ScanHz > TopSpeed then raise Exception.Create(ScanHzEdit.Text + ' Hz intra-scan is too fast.');
    ChPerStart := ChCt * TUInt32(1+Ord(bDualSequencer));
  end
  else begin
    ScanHz := TopSpeed;
    ChPerStart := 1;
  end;

  if Hz * ChPerStart > ScanHz then begin
    if Hz * ChPerStart > TopSpeed then
      raise Exception.Create(HzEdit.Text + ' Hz start rate is too fast.')
    else
      raise Exception.Create(HzEdit.Text + ' Hz start rate is too fast for ' + ScanHzEdit.Text + ' Hz intra-scan.')
    ;
  end;



  case LogFormatRadios.ItemIndex of
    0: LogFormat := lfNone;
    1: LogFormat := lfBinary;
    2: LogFormat := lfSingleSample;
    3: LogFormat := lfAllChannels;
    4: LogFormat := lfClearAll;
    else raise Exception.Create('Invalid log format #' + IntToStr(LogFormatRadios.ItemIndex) + '.');
  end;

  if StopDepthCheck.Checked then
    StopDepth := StrToInt(StopDepthEdit.Text)
  else
    StopDepth := -1
  ;
  DMADataSize := StrToInt('$' + AFEdit.Text);
  if DMADataSize > $FF0 then raise ERangeError.Create('AF/chunk size greater than FF0.');

  Status := DoDMAPiongControl(iBoard, #$10); // AIOAIO.dll function to stop DMA process
  case Status of
    ERROR_SUCCESS: ;
    ERROR_CONNECTION_INVALID: ;
    else begin
      LogMemo.Lines.Add('Error ' + IntToStr(Status) + ' stopping.');
      EnableGo(True);
      Exit;
    end;
  end;

  if DataThread <> nil then begin
    FreeAndNil(DataThread);
  end;

  Sleep(1);

//  BRD_Reset(iBoard);
  PIO_Out32(iBoard, 0, 1); // reset HMAIO and ADIO cards

  Sleep(1);

  PIO_Out32(iBoard, AFLevelOffset, DMADataSize);
  DMADataSize := PIO_In32(iBoard, AFLevelOffset);

  LogMemo.Lines.Add('DMADataSize = ' + IntToHex(DMADataSize, 1));
  if DMADataSize < $100 then begin
    LogMemo.Lines.Add('DMADataSize is too small.');
    EnableGo(True);
    Exit;
  end;

  // Acquire all channels at selected ranges.
  for iCh := 0 to 7 do begin
    ADC_SetRange(iBoard, 0, iCh, RangeRadios[iCh].ItemIndex);
    if bDualSequencer then ADC_SetRange(iBoard, 1, iCh, RangeRadios[iCh+8].ItemIndex);
  end;

  //ADC_ConfigureAcquire(iBoard, iSequencer, ctAuto, 0); //Default.

  //ADC_SetRate(iBoard, 0, Hz);
  BaseClock := PIO_In32(iBoard, BaseClockOffset);
  if bScanMode then begin
    PIO_Out32(iBoard, DivisorOffset + 4, Round(BaseClock / ScanHz));
    PIO_Out32(iBoard, DivisorOffset + 0, Round(BaseClock / Hz));
  end
  else begin
    PIO_Out32(iBoard, DivisorOffset, Round(BaseClock / Hz));
  end;

  // StartStopCmd := StartStopCmd or ($7 shl 12); //All channels.
  StartStopCmd := StartStopCmd or (HighCh shl 12); //All channels.

  if bDifferential then begin
    // Get differential, but only even channels. Odd channels are the low side, IN-, instead of using the common IN-.
    // StartStopCmd := StartStopCmd and not (1 shl 12); //AND out odd channel bit.
    StartStopCmd := StartStopCmd and not (1 shl 11); //AND out single-ended bit.
  end;

  if bScanMode then StartStopCmd := StartStopCmd or (1 shl 18);

  Status := DoDMAPiongControl(iBoard, #$11);   // start DMA
  if Status <> ERROR_SUCCESS then begin
    LogMemo.Lines.Add('Error ' + IntToStr(Status) + AnsiString(' starting.'));
  end;



  { write a new TDataThread Constructor or methods that takes this stuff as parameters ...}
  DataThread := TDataThread.Create;
  DataThread.hDataWnd := Handle;
  DataThread.DataMsg := AM_GOTDATA;
  DataThread.FNSuffix := AnsiString(IntToHex(DeviceID, 4) + ' ' + FloatToStr(Hz)) + ' Hz ';
  if bScanMode then DataThread.FNSuffix := DataThread.FNSuffix + 'Scans ';
  for iCh := $0 to ChCt * TUInt32(1+Ord(bDualSequencer)) - 1 do
    DataThread.FNSuffix := DataThread.FNSuffix + AnsiString(IntToHex(RangeRadios[iCh].ItemIndex, 1))
  ;
  if bDifferential then DataThread.FNSuffix := DataThread.FNSuffix + ' D';
  { ... all of the FNSuffix init stuff should be inside TDataThread methods or constructor }
  DataThread.Start;

  Sleep(1);

  LogMemo.Lines.Add('Start / Stop = ' + IntToHex(StartStopCmd or StartBitMask, 5) + ' / ' + IntToHex(StartStopCmd and not StartBitMask, 5));

  if Model in [HMAIO, ADIO] then begin
    PIO_Out32(iBoard, IRQEnableOffset_ADIO, $00000010);
    if Model = HMAIO then
        PIO_Out32(iBoard, ADCStartStopOffset+$4, StartStopCmd );
    PIO_Out32(iBoard, ADCStartStopOffset, StartStopCmd or StartBitMask);
  end
  else begin // AIO, AIOProto
    PIO_Out8(iBoard, IRQEnableOffset_AIO, $01);
    PIO_Out16(iBoard, ADCStartStopOffset, StartStopCmd or StartBitMask);
  end;

  QueryPerformanceCounter(RunStart);
  LastData := -1;
  LastDataChange := RunStart;
  LastFragmentTicks := RunStart;

  HealthTime.Enabled := True;

  EnableStop(True);
end;

procedure TMainForm.ScanCheckClick(Sender: TObject);
begin
  if ScanCheck.Checked then begin
    ScanGroup.Visible := True;
    HzLabel.Caption := 'Scan Rate(decimal Hz)';
  end
  else begin
    ScanGroup.Visible := False;
    HzLabel.Caption := 'Sample Rate(decimal Hz)';
  end;
  bScanMode := ScanCheck.Checked;
end;

procedure TMainForm.StopButtonClick(Sender: TObject);
var
  Status: TWinErrCode;

begin
  EnableStop(False);

  if StartStopCmd > $FFFF then begin
    PIO_Out32(iBoard, IRQEnableOffset_ADIO, 0);
    //TODO?: if Model = HMAIO do something?
    PIO_Out32(iBoard, ADCStartStopOffset, StartStopCmd and not StartBitMask);
  end
  else begin
    PIO_Out8(iBoard, IRQEnableOffset_AIO, 0);
    PIO_Out16(iBoard, ADCStartStopOffset, StartStopCmd and not StartBitMask);
  end;

  PIO_PollForIRQ(iBoard);

  HealthTime.Enabled := False;

  Sleep(1);

  PIO_PollForIRQ(iBoard);

  //DataList.Items.Clear;
  HealthTime.OnTimer(HealthTime);

  Status := DoDMAPiongControl(iBoard, #$10);
  if Status <> ERROR_SUCCESS then begin
    LogMemo.Lines.Add('Error ' + IntToStr(Status) + ' stopping.');
    EnableGo(True);
    Exit;
  end;

  DataThread.Terminate;
  repeat
    Application.ProcessMessages;
  until DataThread.bDone;
  FreeAndNil(DataThread);

  if Sender <> nil then begin
    RunLength := 60 * TicksPerSec;
    RunsKnown := 0;
  end;

  EnableGo(True);
end;

procedure TMainForm.DoCycleRun(NewHz: Double; ThisRunLength: Int64);
var
  EndTime, dT: Int64;

begin
  StopButtonClick(nil);

  RunLength := RunLength * RunsKnown;
  RunLength := RunLength + ThisRunLength;
  Inc(RunsKnown);
  RunLength := Min(60 * TicksPerSec, RunLength / RunsKnown);

  NewHz := Min(1000000, NewHz);
  NewHz := 1/100 * Round(100 * NewHz);
  HzEdit.Text := FloatToStr(NewHz);

  QueryPerformanceCounter(EndTime);
  EndTime := EndTime + TicksPerSec;
  repeat
    Application.ProcessMessages;
    QueryPerformanceCounter(dT);
    dT := dT - EndTime;
  until dT >= 0;

  GoButtonClick(nil);
end;

procedure TMainForm.ProcessPause;
var
  EndTime, dT: Int64;

begin
  QueryPerformanceCounter(EndTime);
  EndTime := EndTime + TicksPerSec;

  repeat
    Application.ProcessMessages;
    QueryPerformanceCounter(dT);
    dT := dT - EndTime;
  until dT > 0;
end;

procedure TMainForm.AMGotData(var Msg: TMessage);
begin
  Msg.Result := ERROR_SUCCESS;
  while Cardinal(DataList.Items.Count) < Msg.WParam do DataList.Items.Add('');
  DataList.Items[Msg.WParam] := IntToHex(Msg.LParam, 8);
end;

procedure TMainForm.AMMessage(var Msg: TMessage);
var
  pMsg: PAnsiString;

begin
  Msg.Result := ERROR_SUCCESS;
  pMsg := Pointer(Msg.LParam);
  LogMemo.Lines.Add(String(pMsg^));
  Dispose(pMsg);
end;

procedure TMainForm.AMDepth(var Msg: TMessage);
begin
  Msg.Result := ERROR_SUCCESS;
  if Msg.LParam = 0 then
    DepthLabel.Caption := ''
  else
    DepthLabel.Caption := 'Backed Up ' + IntToStr(Msg.LParam)
  ;
  if (StopDepth >= 0) and (Msg.LParam >= StopDepth) and StopButton.Enabled then
    StopButton.Click
  ;
end;

// side effect on Version
function TMainForm.DetectModel(DeviceID: TUInt32): TModelType;
begin
  Version := PIO_In32(iBoard, $68);
  BRD_GetDeviceID(iBoard, @DeviceID);
  if (DeviceID and $4) <> 0 then
  begin
    // if ADIO model (NOTE: ONLY SUPPORTS LATEST ADIO REVISION)
    Result := ADIO;
  end
  else // AIO, now check which
  if (DeviceID and $200) <> 0 then
  begin
    Result := AIOProto;
  end
  else
  begin
    Result := AIO;
    if (Version and $10000) = $10000 then
    begin
      Result := HMAIO;
      caption:='HMAIO';
    end;
  end;
end;

procedure TMainForm.DataListKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
begin
  if (Shift * [ssShift, ssCtrl, ssAlt] = [ssCtrl]) and (Key = $43) then
    Clipboard.AsText := DataList.Items.Text
  ;
end;

function TMainForm.UpdateChannelsAvailable: integer;
var
  I, ChPerSequencer: integer;

begin
  for I := 0 to 15 do
    RangeRadios[I].Visible := False
  ;

  ChPerSequencer := 8;
  if bDifferential then ChPerSequencer := ChPerSequencer div 2;
  ChannelCount.Items.Clear;
  if bDualSequencer then begin
    for I := 0 to ChPerSequencer-1 do begin
      ChannelCount.Items.Add(IntToStr((I+1)*2));
      if bDifferential then begin
        RangeRadios[I*2].Visible := True;
        RangeRadios[I*2+8].Visible := True;
      end
      else begin
        RangeRadios[I].Visible := True;
        RangeRadios[I+8].Visible := True;
      end;
    end;
  end
  else begin
    for I := 0 to ChPerSequencer-1 do begin
      ChannelCount.Items.Add(IntToStr(I+1));
      if bDifferential then
        RangeRadios[I*2].Visible := True
      else
        RangeRadios[I].Visible := True
      ;
    end;
  end;
  for I := 0 to 15 do
    RangeRadios[I].Left := I * RangeRadios[I].Width
  ;
  Result := ChPerSequencer-1;
end;

procedure TMainForm.DeviceComboChange(Sender: TObject);
var
  DeviceID: TUInt32;

begin
  DeviceID := 0;
  iBoard := DeviceCombo.ItemIndex;

  BRD_GetDeviceID(DeviceCombo.ItemIndex, @DeviceID);
  Model := DetectModel(DeviceID);

  if Model in [ADIO] then begin
    RangePanelADAS1.Enabled := False;
    bDualSequencer := False;
  end
  else begin
    RangePanelADAS1.Enabled := True;
    bDualSequencer := True;
  end;
  ChannelCount.ItemIndex := UpdateChannelsAvailable;
end;

procedure TMainForm.HealthTimeTimer(Sender: TObject);
var
  Naui, dT: Int64;
  Buf: String;

begin
  QueryPerformanceCounter(Naui);
  dT := Naui - LastFragmentTicks;

  if dT < TicksPerSec then begin
    dT := LastFragmentTicks - RunStart;
    HealthLabel.Caption := 'Working For ' + IntToStr(Floor(dT / TicksPerSec)) + 's';
  end
  else begin
    Buf := IntToStr(Floor(dT / TicksPerSec)) + 's Ago';
    dT := LastFragmentTicks - RunStart;
    Buf := 'Worked for ' + FloatToStrF(dT / TicksPerSec, ffFixed, 18, 1) + 's, Until ' + Buf;

    HealthLabel.Caption := Buf;
  end;
end;


procedure TMainForm.DifferentialCheckClick(Sender: TObject);
begin
  bDifferential := DifferentialCheck.Checked;
  ChannelCount.ItemIndex := UpdateChannelsAvailable;
end;

end.
